@@ -26,6 +26,7 @@ from itertools import chain
 from operator import attrgetter, itemgetter
 from pathlib import Path
 from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
+from numbers import Number
 
 import torch
 import torchio as tio
@@ -35,7 +36,7 @@ from tqdm.contrib.concurrent import process_map
 from mdlu.data.modality import ImageModality
 from mdlu.utils import PyTorchJsonDecoder, PyTorchJsonEncoder
 
-ImageStats = namedtuple("ImageStats", ("spacing", "spatial_shape", "unique_intensities", "intensity_counts"))
+ImageStats = namedtuple("ImageStats", ("spacing", "spatial_shape", "unique_intensities", "intensity_counts", "num_channels"))
 
 __all__ = ["AbstractDataset", "AbstractDiscreteLabelDataset"]
 
@@ -181,6 +182,7 @@ class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
         "spacings",
         "spatial_shapes",
         "intensity_counts",
+        "num_channels_all_images",
     )
     label_stat_attr_keys: tuple[str, ...] = ()
 
@@ -189,6 +191,7 @@ class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
     spacings: torch.Tensor
     spatial_shapes: torch.Tensor
     intensity_counts: Counter
+    num_channels: torch.Tensor
 
     def __init__(
         self,
@@ -345,7 +348,7 @@ class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
         uniques, counts = image.tensor[image.tensor > image.tensor.min()].unique(return_counts=True)
         uniques, counts = uniques.tolist(), counts.tolist()
 
-        return ImageStats(image.spacing, image.spatial_shape, uniques, counts)
+        return ImageStats(image.spacing, image.spatial_shape, uniques, counts, image.tensor.size(0))
 
     @staticmethod
     def get_single_label_stats(label: Any, *args: Any, **kwargs: Any) -> Any:
@@ -375,6 +378,7 @@ class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
         shapes = tuple(map(attrgetter("spatial_shape"), image_stats))
         unique_intensities = map(attrgetter("unique_intensities"), image_stats)
         intensity_counts = map(attrgetter("intensity_counts"), image_stats)
+        num_channels = tuple(map(attrgetter("num_channels"), image_stats))
 
         for u, c in zip(unique_intensities, intensity_counts):
             intensity_values.update(dict(zip(u, c)))
@@ -383,6 +387,7 @@ class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
             "spacings": torch.tensor(spacings, dtype=torch.float),
             "spatial_shapes": torch.tensor(shapes, dtype=torch.long),
             "intensity_counts": intensity_values,
+            "num_channels_all_images": torch.tensor(num_channels, dtype=torch.long)
         }
 
     @staticmethod
@@ -769,6 +774,15 @@ class AbstractDataset(tio.data.SubjectsDataset, metaclass=ABCMeta):
         """
         return self._target_spacing or self.computed_target_spacing
 
+    @property
+    def num_channels(self) -> int:
+        if self.num_channels_all_images.unique(return_counts=False) == 1:
+            return self.num_channels_all_images[0].item()
+        raise RuntimeError(
+            f"""All images need to have the same number of channels, 
+            but got images with the following numbers of channels: {self.num_channels_all_images}"""
+        )
+
     @property
     def sizes_after_resampling(self) -> torch.Tensor:
         """Computes the image sizes of the dataset after resampling to the median spacing.
@@ -890,10 +904,10 @@ class AbstractDiscreteLabelDataset(AbstractDataset):
     """
 
     class_values: torch.Tensor
-    label_stat_attr_keys = ("class_values", *AbstractDataset.label_stat_attr_keys)
+    label_stat_attr_keys = ("class_values", "spatial_label_shapes", *AbstractDataset.label_stat_attr_keys)
 
     @staticmethod
-    def get_single_label_stats(label: tio.data.Image, *args: Any, **kwargs: Any) -> dict[str, torch.Tensor | Any] | Any:
+    def get_single_label_stats(label: tio.data.Image | torch.Tensor | Number, *args: Any, **kwargs: Any) -> dict[str, torch.Tensor | Any] | Any:
         """Computes the label statistics for a single label.
 
         Args:
@@ -902,7 +916,15 @@ class AbstractDiscreteLabelDataset(AbstractDataset):
         Returns:
             The label statistics for the label.
         """
-        return {"class_values": label.tensor.unique().tolist()}
+        if isinstance(label, tio.data.Image):
+            label_tensor = label.tensor
+        elif isinstance(label, torch.Tensor):
+            label_tensor = label
+        elif isinstance(label, Number):
+            label_tensor = torch.tensor(label).view(1,)
+        else:
+            raise TypeError(f'label has to be either torchio Image, torch.Tensor or Number, got {type(label)}.')
+        return {"class_values": label_tensor.unique().tolist(), 'spatial_shape': label_tensor.shape[1:]}
 
     @staticmethod
     def aggregate_label_stats(*label_stats: Any) -> dict[str, torch.Tensor | Any] | None:
@@ -915,9 +937,20 @@ class AbstractDiscreteLabelDataset(AbstractDataset):
             The aggregated label statistics.
         """
         return {
-            "class_values": torch.tensor(sorted(set(chain.from_iterable(map(itemgetter("class_values"), label_stats)))))
+            "class_values": torch.tensor(sorted(set(chain.from_iterable(map(itemgetter("class_values"), label_stats))))),
+            "spatial_label_shapes": torch.tensor(tuple(map(attrgetter("spatial_shape"), image_stats)), dtype=torch.long)
         }
 
+    @property
+    def target_label_size(self) -> torch.Tensor:
+        if torch.allclose(self.spatial_label_shapes, self.spatial_shapes):
+            return self.target_size
+        
+        median = spatial_shapes.median(0)
+        if torch.isnan(median).any():
+            return torch.tensor([1])
+        return median
+
     @property
     def consecutive_class_mapping(self) -> dict[int, int]:
         """Returns the mapping from consecutive class values to the original class values.
